#include <iostream>
#include <fstream>
#include <vector>
#include <Eigen/Dense>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include "WatsonNadaraya.hh"
#include "svm_common.hh"
using namespace boost::numeric::ublas;


#if defined WIN32 || WIN64 || _WIN32 || _WIN64
#include "getopt.h"
#else 
#include <unistd.h>
#endif 

static int verbose_flag;
static int train_flag;
static int test_flag;

static char help_msg[] = 
"Usage: runWN [-d TrainFileName] [-t TestFileName] [-h] [--verbose]\n\
  -h [ --help ]: print this help.\n\
  -d [ --train ] TrainFileName: set the training input file.\n\
  -t [ --test ] TestFileName: set the testing input file.\n\
  [ --verbose ]: verbose flag.\n\
";

int main(int argc, char** argv)
{
    int c;
    std::string trainFile;
    std::string testFile;
    //std::string oFile;
    static struct option long_options[]=
    {
        {"verbose", no_argument, &verbose_flag, 1},
        {"train", required_argument, 0, 'd'},
        {"test", required_argument, 0, 't'},
        {0,0,0,0}
    };

    int option_index=0;
    
    int errflg=0;
    while ((c = getopt_long(argc, argv, "d:t:h",long_options,&option_index)) != -1)
    {
        switch (c) {
        case 0:
            break;
        case 'd':
            trainFile=std::string(optarg);
            train_flag=1;
            break;
        case 't':
            test_flag=1;
            testFile=std::string(optarg);
            break;
        case '?':
            errflg++;
            break;
        default:
            errflg++;
            break;
        }
        
    }
    /*ConfigFile config;
    try
    {
        config=ConfigFile(configfile.c_str());
    }
    catch (ConfigFile::file_not_found)
    {
        errflg++;
        std::cerr << configfile << " not found!" << std::endl;
    }*/
    if (errflg || argc == 1) {
        std::cout << help_msg << std::endl;
        exit (0);
    }
    if (train_flag && test_flag)
    {
        by::SvmDoc sdoc(trainFile,1);
        vector<int> x_label(sdoc.totdoc);
        matrix<double> x_train(sdoc.totdoc,sdoc.totwords);
        for (int i=0; i<sdoc.totdoc; i++)
        {
            x_train(i,0)=sdoc.fvec[i].words[0].weight;
            x_label(i) = (sdoc.label[i]+1)/2;
        }
        by::SvmDoc tdoc(testFile,1);
        matrix<double> x_test(tdoc.totdoc,tdoc.totwords);
        for (int i=0; i<tdoc.totdoc; i++)
        {
            x_test(i,0)=tdoc.fvec[i].words[0].weight;
            tdoc.label[i] = (tdoc.label[i]+1)/2;
        }
        by::WatsonNadaraya wnclassifier(2,"knn");
        
        vector<int> y_test = wnclassifier.classify(x_train, x_label, x_test);
        int numWrong = 0;
        for (int i=0; i<tdoc.totdoc; i++)
        {
            if (y_test(i) != tdoc.label[i]) numWrong++;
        }
        std::cout << "Numwrong: " << numWrong << "/" << tdoc.totdoc << std::endl;
    }
    else{
        //TODO;
    }
}
